/*
 * Copyright (c) 2005-2007 Obeo
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package fr.obeo.acceleo.gen.template;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.Path;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.ui.texteditor.MarkerUtilities;

import fr.obeo.acceleo.ecore.factories.FactoryException;
import fr.obeo.acceleo.gen.AcceleoEcoreGenPlugin;
import fr.obeo.acceleo.gen.template.debug.ITemplateDebugger;
import fr.obeo.acceleo.gen.template.debug.TemplateDebugger;
import fr.obeo.acceleo.gen.template.eval.ENode;
import fr.obeo.acceleo.gen.template.eval.ENodeCastException;
import fr.obeo.acceleo.gen.template.eval.ENodeException;
import fr.obeo.acceleo.gen.template.eval.LaunchManager;
import fr.obeo.acceleo.gen.template.scripts.IScript;
import fr.obeo.acceleo.gen.template.statements.TemplateFeatureStatement;
import fr.obeo.acceleo.tools.resources.Resources;
import fr.obeo.acceleo.tools.strings.Int2;
import fr.obeo.acceleo.tools.strings.TextSearch;

/**
 * This is a template's element for the generation tool.
 * 
 * @author www.obeo.fr
 * 
 */
public abstract class TemplateElement {

	/**
	 * The script.
	 */
	protected IScript script;

	/**
	 * The parent element.
	 */
	protected TemplateElement parent = null;

	/**
	 * The children.
	 */
	protected List children = new ArrayList();

	/**
	 * The next element.
	 */
	protected TemplateElement next = null;

	/**
	 * The previous element.
	 */
	protected TemplateElement previous = null;

	/**
	 * The position in the script file.
	 */
	protected Int2 pos = Int2.NOT_FOUND;

	/**
	 * The line in the script file.
	 */
	protected Integer line = null;

	/**
	 * Debugger for all template evaluation.
	 */
	private static TemplateDebugger debugger = null;

	/**
	 * Gets the debugger for all template evaluation.
	 */
	public static ITemplateDebugger getDebugger() {
		if (debugger == null) {
			debugger = new TemplateDebugger();
		}
		return debugger;
	}

	/**
	 * Constructor.
	 * 
	 * @param script
	 *            is the script
	 */
	public TemplateElement(IScript script) {
		this.script = script;
	}

	/**
	 * @return the script
	 */
	public IScript getScript() {
		return script;
	}

	/**
	 * @return the position of this expression
	 */
	public Int2 getPos() {
		return this.pos;
	}

	/**
	 * @param pos
	 *            is the position of this expression
	 */
	public void setPos(Int2 pos) {
		this.pos = pos;
		line = null;
	}

	/**
	 * @return the line of this expression
	 */
	public int getLine() {
		if (line == null) {
			if (pos.b() == -1 || getScript() == null || getScript().getFile() == null) {
				line = new Integer(0);
			} else {
				line = new Integer(TextSearch.getDefaultSearch().lineNumber(getScript().getFile(), pos.b()));
			}
		}
		return line.intValue();
	}

	/**
	 * Sets the parent element. This element is a child for the parent.
	 * 
	 * @param parent
	 *            is the parent element
	 */
	public void setParent(TemplateElement parent) {
		// ASSERT this.parent == null
		this.parent = parent;
		if (parent != null) {
			if (parent.children.size() > 0) {
				TemplateElement previous = (TemplateElement) parent.children.get(parent.children.size() - 1);
				this.previous = previous;
				previous.next = this;
			}
			parent.children.add(this);
		}
	}

	/**
	 * @return the parent element
	 */
	public TemplateElement getParent() {
		return parent;
	}

	/**
	 * @return the previous element
	 */
	public TemplateElement getPrevious() {
		return previous;
	}

	/**
	 * @return the next element
	 */
	public TemplateElement getNext() {
		return next;
	}

	/**
	 * Returns a table of children.
	 * 
	 * @return a table of children
	 */
	public TemplateElement[] getChildren() {
		return (TemplateElement[]) children.toArray(new TemplateElement[children.size()]);
	}

	/**
	 * Returns a table of the children that are instance of the given class.
	 * 
	 * @param c
	 *            is a class
	 * @return a table of the children that are instance of the given class
	 */
	public TemplateElement[] getChildren(Class c) {
		List result = new ArrayList();
		Iterator it = children.iterator();
		while (it.hasNext()) {
			TemplateElement element = (TemplateElement) it.next();
			if (c.isInstance(element)) {
				result.add(element);
			}
		}
		return (TemplateElement[]) result.toArray(new TemplateElement[result.size()]);
	}

	/**
	 * Returns a table of the children and the gateway children that are
	 * instance of the given classes.
	 * 
	 * @param classes
	 *            are the kept children
	 * @param gatewayClasses
	 *            are the gateway nodes that aren't kept, but their children are
	 *            tested
	 * @return a table of the children that are instance of the given classes
	 */
	public TemplateElement[] getChildren(Class[] classes, Class[] gatewayClasses) {
		List result = new ArrayList();
		Iterator it = children.iterator();
		while (it.hasNext()) {
			TemplateElement element = (TemplateElement) it.next();
			boolean instance = false;
			for (int i = 0; i < gatewayClasses.length && !instance; i++) {
				Class c = gatewayClasses[i];
				if (c.isInstance(element)) {
					result.addAll(Arrays.asList(element.getChildren(classes, gatewayClasses)));
					instance = true;
				}
			}
			for (int i = 0; i < classes.length && !instance; i++) {
				Class c = classes[i];
				if (c.isInstance(element)) {
					result.add(element);
					instance = true;
				}
			}
		}
		return (TemplateElement[]) result.toArray(new TemplateElement[result.size()]);
	}

	/**
	 * Returns a list of all the children that are instance of the given class.
	 * The result contains also the children of the children...
	 * 
	 * @param c
	 *            is a class
	 * @return a list of all the children that are instance of the given class
	 */
	public List getAllElements(Class c) {
		List result = new ArrayList();
		if (c.isInstance(this)) {
			result.add(this);
		}
		Iterator it = children.iterator();
		while (it.hasNext()) {
			TemplateElement element = (TemplateElement) it.next();
			result.addAll(element.getAllElements(c));
		}
		return result;
	}

	/**
	 * Returns the URI fragment of this template element.
	 * 
	 * @return the URI fragment of this template element
	 */
	public String getURIFragment() {
		StringBuffer fragment = new StringBuffer("");
		if (script != null && script.getFile() != null) {
			fragment.append(Resources.encodeAcceleoAbsolutePath(script.getFile().getAbsolutePath()));
		}
		fragment.append(" //pos=");
		fragment.append(pos.b());
		fragment.append(",");
		fragment.append(pos.e());
		return fragment.toString();
	}

	/**
	 * Gets the file that corresponds to the given fragment.
	 * 
	 * @param uriFragment
	 *            is the URI of the template element
	 * @return the file, or null if it doesn't exist
	 */
	public static File getTemplateFileFromURIFragment(String uriFragment) {
		if (uriFragment != null) {
			String path;
			int i = uriFragment.indexOf(" //pos=");
			if (i > -1) {
				path = uriFragment.substring(0, i).trim();
			} else {
				path = uriFragment;
			}
			File file = new File(Resources.decodeAcceleoAbsolutePath(path));
			if (file.exists()) {
				return file;
			}
		}
		return null;
	}

	/**
	 * Gets the position that corresponds to the given fragment.
	 * 
	 * @param uriFragment
	 *            is the URI of the template element
	 * @return the position in the template file, or Int2.NotFound if it doesn't
	 *         exist
	 */
	public static Int2 getPositionFromURIFragment(String uriFragment) {
		if (uriFragment != null) {
			String search = " //pos=";
			int i = uriFragment.indexOf(search);
			if (i > -1) {
				String pos = uriFragment.substring(i + search.length()).trim();
				i = pos.indexOf(",");
				if (i > -1) {
					String b = pos.substring(0, i).trim();
					String e = pos.substring(i + 1).trim();
					return new Int2(Integer.parseInt(b), Integer.parseInt(e));
				}
			}
		}
		return Int2.NOT_FOUND;
	}

	/**
	 * Starts the debug action.
	 * 
	 * @param mode
	 *            is the mode
	 */
	protected void startDebug(LaunchManager mode) {
		if (mode.getMode() == LaunchManager.DEBUG_MODE) {
			debugger.pushStack(getScript().getFile(), getLine(), getPos());
		}
	}

	/**
	 * Updates the current input of the debugger.
	 * 
	 * @param mode
	 *            is the mode
	 * @param input
	 *            is the new input
	 */
	protected void stepDebugInput(LaunchManager mode, Object input) {
		if (mode.getMode() == LaunchManager.DEBUG_MODE) {
			if (debugger.isBreakpoint(getScript().getFile(), getLine(), getPos()) && conditionOK(getScript().getFile(), getLine(), input)) {
				Map map = getInputVariables(input);
				debugger.waitForEvent(map);
			} else {
				Map map = getInputVariables(input);
				debugger.updateVariables(map);
			}
		}
	}

	private Map getInputVariables(Object input) {
		Map map = new HashMap();
		map.put("input", input);
		if (script != null) {
			map.put("i", script.contextPeek(IScript.WHILE_INDEX));
			List currents = new ArrayList();
			int i = 0;
			Object iObject = script.contextAt(IScript.TEMPLATE_NODE, i);
			while (iObject != null) {
				currents.add(iObject);
				i++;
				iObject = script.contextAt(IScript.TEMPLATE_NODE, i);
			}
			map.put("current(i)", currents);
		}
		return map;
	}

	/**
	 * Updates the current output of the debugger.
	 * 
	 * @param mode
	 *            is the mode
	 * @param output
	 *            is the new output
	 */
	protected void stepDebugOutput(LaunchManager mode, Object input, ENode output) {
		if (mode.getMode() == LaunchManager.DEBUG_MODE) {
			if (debugger.isBreakpoint(getScript().getFile(), getLine(), getPos()) && conditionOK(getScript().getFile(), getLine(), input)) {
				Map map = new HashMap();
				map.put("output", output);
				debugger.waitForEvent(map);
			}
		}
	}

	private boolean conditionOK(File file, int line, Object input) {
		IFile workspaceFile = ResourcesPlugin.getWorkspace().getRoot().getFileForLocation(new Path(file.getAbsolutePath()));
		if (workspaceFile != null && workspaceFile.isAccessible()) {
			String condition = null;
			try {
				IMarker[] markers = workspaceFile.findMarkers("org.eclipse.debug.core.lineBreakpointMarker", true, 1);
				for (int i = 0; condition == null && i < markers.length; i++) {
					if (MarkerUtilities.getLineNumber(markers[i]) == line) {
						condition = markers[i].getAttribute("condition", "");
					}
				}
			} catch (CoreException e) {
				AcceleoEcoreGenPlugin.getDefault().log(e, true);
			}
			if (condition != null && condition.trim().length() > 0) {
				Template template = new Template(getScript());
				try {
					template.append(TemplateFeatureStatement.fromString(condition, new Int2(0, condition.length()), getScript()));
				} catch (TemplateSyntaxException e) {
					AcceleoEcoreGenPlugin.getDefault().log(e, true);
					return false;
				}
				ENodeException.disableRuntimeMarkersFor(this);
				try {
					if (input instanceof ENode) {
						input = ((ENode) input).getAdapterValue(EObject.class);
					}
					if (input instanceof EObject) {
						ENode result = template.evaluate((EObject) input, LaunchManager.create("run", false));
						return ((Boolean) result.getAdapterValue(boolean.class)).booleanValue();
					} else {
						return false;
					}
				} catch (ENodeException e) {
					return false;
				} catch (FactoryException e) {
					return false;
				} catch (ENodeCastException e) {
					return false;
				} finally {
					ENodeException.enableRuntimeMarkersFor(this);
				}
			} else {
				return true;
			}
		} else {
			return true;
		}
	}

	/**
	 * Stops the debug action.
	 * 
	 * @param mode
	 *            is the mode
	 */
	protected void endDebug(LaunchManager mode) {
		if (mode.getMode() == LaunchManager.DEBUG_MODE) {
			debugger.popStack();
		}
	}

}