/*
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    David Bonneau and Obeo - initial API and implementation
 */

package fr.obeo.acceleo.gen.debug.ui.model;

import java.util.Iterator;

import org.eclipse.debug.core.DebugException;
import org.eclipse.debug.core.model.IRegisterGroup;
import org.eclipse.debug.core.model.IStackFrame;
import org.eclipse.debug.core.model.IThread;
import org.eclipse.debug.core.model.IVariable;

import fr.obeo.acceleo.gen.template.debug.StackInfo;

/**
 * Acceleo stack frame.
 */
public class AcceleoStackFrame extends AcceleoDebugElement implements IStackFrame {

	private AcceleoThread thread;

	private StackInfo stackInfo;

	/**
	 * Constructs a stack frame in the given thread with the given frame data.
	 * 
	 * @param thread
	 * @param data
	 *            frame data
	 * @param id
	 *            stack frame id (0 is the bottom of the stack)
	 */
	public AcceleoStackFrame(AcceleoThread thread, StackInfo aStackInfo) {
		super((AcceleoDebugTarget) thread.getDebugTarget());
		this.thread = thread;
		stackInfo = aStackInfo;
	}

	/* (non-Javadoc) */
	public IThread getThread() {
		return thread;
	}

	/* (non-Javadoc) */
	public IVariable[] getVariables() throws DebugException {
		IVariable[] ret = new AcceleoVariable[stackInfo.getVariables().keySet().size()];
		int i = 0;
		for (Iterator iterator = stackInfo.getVariables().keySet().iterator(); iterator.hasNext();) {
			String name = (String) iterator.next();
			Object value = (Object) stackInfo.getVariables().get(name);
			ret[i] = new AcceleoVariable(this, name, value, AcceleoVariable.DEFAULT_TYPE);
			++i;
		}
		return ret;
	}

	/* (non-Javadoc) */
	public boolean hasVariables() throws DebugException {
		return getVariables().length > 0;
	}

	/* (non-Javadoc) */
	public int getLineNumber() throws DebugException {
		return stackInfo.getLine();
	}

	/* (non-Javadoc) */
	public int getCharStart() throws DebugException {
		return stackInfo.getCharStart();
	}

	/* (non-Javadoc) */
	public int getCharEnd() throws DebugException {
		return stackInfo.getCharEnd();
	}

	/* (non-Javadoc) */
	public String getName() throws DebugException {
		return stackInfo.getFile().getName();
	}

	/* (non-Javadoc) */
	public IRegisterGroup[] getRegisterGroups() throws DebugException {
		return null;
	}

	/* (non-Javadoc) */
	public boolean hasRegisterGroups() throws DebugException {
		return false;
	}

	/* (non-Javadoc) */
	public boolean canStepInto() {
		return getThread().canStepInto();
	}

	/* (non-Javadoc) */
	public boolean canStepOver() {
		return getThread().canStepOver();
	}

	/* (non-Javadoc) */
	public boolean canStepReturn() {
		return getThread().canStepReturn();
	}

	/* (non-Javadoc) */
	public boolean isStepping() {
		return getThread().isStepping();
	}

	/* (non-Javadoc) */
	public void stepInto() throws DebugException {
		getThread().stepInto();
	}

	/* (non-Javadoc) */
	public void stepOver() throws DebugException {
		getThread().stepOver();
	}

	/* (non-Javadoc) */
	public void stepReturn() throws DebugException {
		getThread().stepReturn();
	}

	/* (non-Javadoc) */
	public boolean canResume() {
		return getThread().canResume();
	}

	/* (non-Javadoc) */
	public boolean canSuspend() {
		return getThread().canSuspend();
	}

	/* (non-Javadoc) */
	public boolean isSuspended() {
		return getThread().isSuspended();
	}

	/* (non-Javadoc) */
	public void resume() throws DebugException {
		getThread().resume();
	}

	/* (non-Javadoc) */
	public void suspend() throws DebugException {
		getThread().suspend();
	}

	/* (non-Javadoc) */
	public boolean canTerminate() {
		return getThread().canTerminate();
	}

	/* (non-Javadoc) */
	public boolean isTerminated() {
		return getThread().isTerminated();
	}

	/* (non-Javadoc) */
	public void terminate() throws DebugException {
		getThread().terminate();
	}

	/**
	 * Returns the name of the source file this stack frame is associated with.
	 * 
	 * @return the name of the source file this stack frame is associated with
	 */
	public String getSourceName() {
		return stackInfo.getFile().getName();
	}

	/* (non-Javadoc) */
	public boolean equals(Object obj) {
		if (obj instanceof AcceleoStackFrame) {
			AcceleoStackFrame sf = (AcceleoStackFrame) obj;
			try {
				return sf.getSourceName().equals(getSourceName()) && sf.getLineNumber() == getLineNumber();
			} catch (DebugException e) {
			}
		}
		return false;
	}

	/* (non-Javadoc) */
	public int hashCode() {
		return getSourceName().hashCode();
	}

	/**
	 * Returns the debugger execution context.
	 * 
	 * @return the debugger execution context
	 */
	public StackInfo getStackInfo() {
		return stackInfo;
	}

}
